locals {
  #Forms the list of resources for the dynamo read-write policy as it is conditioned to the stream_enabled variable
  policy_resources = var.stream_enabled ? "\"${aws_dynamodb_table.table.arn}\",\"${aws_dynamodb_table.table.stream_arn}\",\"${aws_dynamodb_table.table.arn}/index/*\"" : "\"${aws_dynamodb_table.table.arn}\",\"${aws_dynamodb_table.table.arn}/index/*\""
}

# This chart only creates a single table
resource "aws_dynamodb_table" "table" {
  name             = var.table_name
  billing_mode     = var.billing_mode
  read_capacity    = var.read_capacity
  write_capacity   = var.write_capacity
  hash_key         = var.hash_key
  range_key        = var.range_key
  stream_enabled   = var.stream_enabled
  stream_view_type = var.stream_view_type

  # String attributes
  dynamic "attribute" {
    for_each = var.string_attributes
    content {
      name = attribute.value
      type = "S"
    }
  }

  # Number attributes
  dynamic "attribute" {
    for_each = var.number_attributes
    content {
      name = attribute.value
      type = "N"
    }
  }

  # Binary attributes
  dynamic "attribute" {
    for_each = var.binary_attributes
    content {
      name = attribute.value
      type = "B"
    }
  }


  # TODO - do we want to expose local_secondary_index as well?

  dynamic "global_secondary_index" {
    for_each = var.global_secondary_index
    content {
      name               = global_secondary_index.value.name
      hash_key           = global_secondary_index.value.hash_key
      projection_type    = global_secondary_index.value.projection_type
      range_key          = lookup(global_secondary_index.value, "range_key", null)
      read_capacity      = lookup(global_secondary_index.value, "read_capacity", null)
      write_capacity     = lookup(global_secondary_index.value, "write_capacity", null)
      non_key_attributes = lookup(global_secondary_index.value, "non_key_attributes", [])
    }
  }

  dynamic "ttl" {
    for_each = length(var.ttl_attribute_name) > 0 ? ["true"] : []
    content {
      attribute_name = var.ttl_attribute_name
      enabled        = true
    }
  }

  tags = var.tags

  point_in_time_recovery {
    enabled = var.point_in_time_recovery_enabled
  }

  # Use the AWS managed CMK
  server_side_encryption {
    enabled = var.enable_sse
  }

  # This is REQUIRED when using Application Autoscaling - otherwise Terraform Applies will overwrite the autoscaled capacity.
  lifecycle {
    ignore_changes = [
      read_capacity,
      write_capacity,
      # We are unable to ignore just the capacity attributes of a global_secondary_index.
      # This is also where our restriction to never change existing GSIs originates.
      # See https://github.com/hashicorp/terraform-provider-aws/issues/671 for more details.
      global_secondary_index,
    ]
  }
}

# Autoscaling for hash key
resource "aws_appautoscaling_target" "dynamodb_table_read_target" {
  count              = var.billing_mode == "PROVISIONED" ? 1 : 0
  max_capacity       = var.max_read_capacity
  min_capacity       = var.min_read_capacity
  resource_id        = "table/${var.table_name}"
  scalable_dimension = "dynamodb:table:ReadCapacityUnits"
  service_namespace  = "dynamodb"

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

resource "aws_appautoscaling_policy" "dynamodb_table_read_policy" {
  count              = var.billing_mode == "PROVISIONED" ? 1 : 0
  name               = "DynamoDBReadCapacityUtilization:${aws_appautoscaling_target.dynamodb_table_read_target[0].resource_id}"
  policy_type        = "TargetTrackingScaling"
  resource_id        = aws_appautoscaling_target.dynamodb_table_read_target[0].resource_id
  scalable_dimension = aws_appautoscaling_target.dynamodb_table_read_target[0].scalable_dimension
  service_namespace  = aws_appautoscaling_target.dynamodb_table_read_target[0].service_namespace

  target_tracking_scaling_policy_configuration {
    predefined_metric_specification {
      predefined_metric_type = "DynamoDBReadCapacityUtilization"
    }

    target_value       = var.read_capacity_utilization_target
    scale_in_cooldown  = 0
    scale_out_cooldown = 0
  }

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

resource "aws_appautoscaling_target" "dynamodb_table_write_target" {
  count              = var.billing_mode == "PROVISIONED" ? 1 : 0
  max_capacity       = var.max_write_capacity
  min_capacity       = var.min_write_capacity
  resource_id        = "table/${var.table_name}"
  scalable_dimension = "dynamodb:table:WriteCapacityUnits"
  service_namespace  = "dynamodb"

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

resource "aws_appautoscaling_policy" "dynamodb_table_write_policy" {
  count              = var.billing_mode == "PROVISIONED" ? 1 : 0
  name               = "DynamoDBwriteCapacityUtilization:${aws_appautoscaling_target.dynamodb_table_write_target[0].resource_id}"
  policy_type        = "TargetTrackingScaling"
  resource_id        = aws_appautoscaling_target.dynamodb_table_write_target[0].resource_id
  scalable_dimension = aws_appautoscaling_target.dynamodb_table_write_target[0].scalable_dimension
  service_namespace  = aws_appautoscaling_target.dynamodb_table_write_target[0].service_namespace

  target_tracking_scaling_policy_configuration {
    predefined_metric_specification {
      predefined_metric_type = "DynamoDBWriteCapacityUtilization"
    }

    target_value       = var.write_capacity_utilization_target
    scale_in_cooldown  = 0
    scale_out_cooldown = 0
  }

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

# Autoscaling for indexes
resource "aws_appautoscaling_target" "dynamodb_index_read_target" {
  for_each           = var.autoscaling_indexes
  max_capacity       = each.value["max_read_capacity"]
  min_capacity       = each.value["min_read_capacity"]
  resource_id        = "table/${var.table_name}/index/${replace(format("%s", each.key), "_", "-")}"
  scalable_dimension = "dynamodb:index:ReadCapacityUnits"
  service_namespace  = "dynamodb"

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

resource "aws_appautoscaling_policy" "dynamodb_index_read_policy" {
  for_each           = var.autoscaling_indexes
  name               = "DynamoDBReadCapacityUtilization:${aws_appautoscaling_target.dynamodb_index_read_target[each.key].resource_id}"
  policy_type        = "TargetTrackingScaling"
  resource_id        = aws_appautoscaling_target.dynamodb_index_read_target[each.key].resource_id
  scalable_dimension = aws_appautoscaling_target.dynamodb_index_read_target[each.key].scalable_dimension
  service_namespace  = aws_appautoscaling_target.dynamodb_index_read_target[each.key].service_namespace

  target_tracking_scaling_policy_configuration {
    predefined_metric_specification {
      predefined_metric_type = "DynamoDBReadCapacityUtilization"
    }

    target_value       = each.value["read_utilization_target"]
    scale_in_cooldown  = 0
    scale_out_cooldown = 0
  }

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

resource "aws_appautoscaling_target" "dynamodb_index_write_target" {
  for_each           = var.autoscaling_indexes
  max_capacity       = each.value["max_write_capacity"]
  min_capacity       = each.value["min_write_capacity"]
  resource_id        = "table/${var.table_name}/index/${replace(format("%s", each.key), "_", "-")}"
  scalable_dimension = "dynamodb:index:WriteCapacityUnits"
  service_namespace  = "dynamodb"

  depends_on = [
    aws_dynamodb_table.table,
  ]
}

resource "aws_appautoscaling_policy" "dynamodb_index_write_policy" {
  for_each           = var.autoscaling_indexes
  name               = "DynamoDBwriteCapacityUtilization:${aws_appautoscaling_target.dynamodb_index_write_target[each.key].resource_id}"
  policy_type        = "TargetTrackingScaling"
  resource_id        = aws_appautoscaling_target.dynamodb_index_write_target[each.key].resource_id
  scalable_dimension = aws_appautoscaling_target.dynamodb_index_write_target[each.key].scalable_dimension
  service_namespace  = aws_appautoscaling_target.dynamodb_index_write_target[each.key].service_namespace

  target_tracking_scaling_policy_configuration {
    predefined_metric_specification {
      predefined_metric_type = "DynamoDBWriteCapacityUtilization"
    }

    target_value       = each.value["write_utilization_target"]
    scale_in_cooldown  = 0
    scale_out_cooldown = 0
  }

  depends_on = [
    aws_dynamodb_table.table,
  ]
}


resource "aws_iam_policy" "read-only-policy" {
  count       = var.enable_iam_policies ? 1 : 0
  name        = "${var.iam_policy_prefix}-dynamodb-read-only"
  path        = "/"
  description = "DynamoDB read-only policy"

  policy = <<EOF
{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Action": [
        "dynamodb:BatchGetItem",
        "dynamodb:GetItem",
        "dynamodb:DescribeTable",
        "dynamodb:Query",
        "dynamodb:Scan",
        "dynamodb:ConditionCheckItem"
      ],
      "Effect": "Allow",
      "Resource": "${aws_dynamodb_table.table.arn}"
    }
  ]
}
EOF
  lifecycle {
    ignore_changes = [
      description,
    ]
  }
  tags = var.tags
}

resource "aws_iam_policy" "read-write-policy" {
  count       = var.enable_iam_policies ? 1 : 0
  name        = "${var.iam_policy_prefix}-dynamodb-read-write"
  path        = "/"
  description = "DynamoDB read and write policy"

  policy = <<EOF
{
  "Version": "2012-10-17",
  "Statement": [
    {
      "Action": [
        "dynamodb:DescribeTable",
        "dynamodb:DescribeStream",
        "dynamodb:GetRecords",
        "dynamodb:GetShardIterator",
        "dynamodb:ListStreams",
        "dynamodb:Query",
        "dynamodb:PutItem",
        "dynamodb:GetItem",
        "dynamodb:Scan",
        "dynamodb:BatchWriteItem",
        "dynamodb:UpdateItem",
        "dynamodb:BatchGetItem",
        "dynamodb:DeleteItem"
      ],
      "Effect": "Allow",
      "Resource": [${local.policy_resources}]
    }
  ]
}
EOF
  lifecycle {
    ignore_changes = [
      description,
    ]
  }
  tags = var.tags
}
